"""Define collections of generic key-bindings which do not belong to widgets."""

from euporie.core.key_binding.bindings import basic, completion, micro, mouse

__all__ = ["basic", "completion", "micro", "mouse"]
